{
  "type": "record",
  "name": "GlossaryNodeKey",
  "namespace": "com.linkedin.pegasus2avro.metadata.key",
  "doc": "Key for a GlossaryNode",
  "fields": [
    {
      "name": "name",
      "type": "string",
      "Searchable": {
        "enableAutocomplete": true,
        "fieldType": "TEXT_PARTIAL"
      }
    }
  ],
  "Aspect": {
    "name": "glossaryNodeKey",
    "keyForEntity": "glossaryNode",
    "entityCategory": "_unset_",
    "entityAspects": [
      "glossaryNodeInfo",
      "institutionalMemory",
      "ownership",
      "status"
    ]
  }
}