# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off

import functools
import pathlib

def _load_schema(schema_name: str) -> str:
    return (pathlib.Path(__file__).parent / f"{schema_name}.avsc").read_text()

@functools.lru_cache(maxsize=None)
def getTelemetryClientIdSchema() -> str:
    return _load_schema("TelemetryClientId")

@functools.lru_cache(maxsize=None)
def getDataProcessInstanceRelationshipsSchema() -> str:
    return _load_schema("DataProcessInstanceRelationships")

@functools.lru_cache(maxsize=None)
def getDataProcessInstancePropertiesSchema() -> str:
    return _load_schema("DataProcessInstanceProperties")

@functools.lru_cache(maxsize=None)
def getDataProcessInstanceOutputSchema() -> str:
    return _load_schema("DataProcessInstanceOutput")

@functools.lru_cache(maxsize=None)
def getDataProcessInstanceInputSchema() -> str:
    return _load_schema("DataProcessInstanceInput")

@functools.lru_cache(maxsize=None)
def getDataProcessInfoSchema() -> str:
    return _load_schema("DataProcessInfo")

@functools.lru_cache(maxsize=None)
def getDataProcessInstanceRunEventSchema() -> str:
    return _load_schema("DataProcessInstanceRunEvent")

@functools.lru_cache(maxsize=None)
def getAssertionInfoSchema() -> str:
    return _load_schema("AssertionInfo")

@functools.lru_cache(maxsize=None)
def getAssertionRunEventSchema() -> str:
    return _load_schema("AssertionRunEvent")

@functools.lru_cache(maxsize=None)
def getGlossaryTermInfoSchema() -> str:
    return _load_schema("GlossaryTermInfo")

@functools.lru_cache(maxsize=None)
def getGlossaryRelatedTermsSchema() -> str:
    return _load_schema("GlossaryRelatedTerms")

@functools.lru_cache(maxsize=None)
def getGlossaryNodeInfoSchema() -> str:
    return _load_schema("GlossaryNodeInfo")

@functools.lru_cache(maxsize=None)
def getExecutionRequestResultSchema() -> str:
    return _load_schema("ExecutionRequestResult")

@functools.lru_cache(maxsize=None)
def getExecutionRequestSignalSchema() -> str:
    return _load_schema("ExecutionRequestSignal")

@functools.lru_cache(maxsize=None)
def getExecutionRequestInputSchema() -> str:
    return _load_schema("ExecutionRequestInput")

@functools.lru_cache(maxsize=None)
def getMetadataChangeProposalSchema() -> str:
    return _load_schema("MetadataChangeProposal")

@functools.lru_cache(maxsize=None)
def getMetadataChangeEventSchema() -> str:
    return _load_schema("MetadataChangeEvent")

@functools.lru_cache(maxsize=None)
def getMetadataChangeLogSchema() -> str:
    return _load_schema("MetadataChangeLog")

@functools.lru_cache(maxsize=None)
def getPlatformEventSchema() -> str:
    return _load_schema("PlatformEvent")

@functools.lru_cache(maxsize=None)
def getDatasetProfileSchema() -> str:
    return _load_schema("DatasetProfile")

@functools.lru_cache(maxsize=None)
def getViewPropertiesSchema() -> str:
    return _load_schema("ViewProperties")

@functools.lru_cache(maxsize=None)
def getEditableDatasetPropertiesSchema() -> str:
    return _load_schema("EditableDatasetProperties")

@functools.lru_cache(maxsize=None)
def getDatasetUpstreamLineageSchema() -> str:
    return _load_schema("DatasetUpstreamLineage")

@functools.lru_cache(maxsize=None)
def getDatasetDeprecationSchema() -> str:
    return _load_schema("DatasetDeprecation")

@functools.lru_cache(maxsize=None)
def getDatasetPropertiesSchema() -> str:
    return _load_schema("DatasetProperties")

@functools.lru_cache(maxsize=None)
def getDatasetUsageStatisticsSchema() -> str:
    return _load_schema("DatasetUsageStatistics")

@functools.lru_cache(maxsize=None)
def getUpstreamLineageSchema() -> str:
    return _load_schema("UpstreamLineage")

@functools.lru_cache(maxsize=None)
def getEditableDashboardPropertiesSchema() -> str:
    return _load_schema("EditableDashboardProperties")

@functools.lru_cache(maxsize=None)
def getDashboardInfoSchema() -> str:
    return _load_schema("DashboardInfo")

@functools.lru_cache(maxsize=None)
def getDashboardUsageStatisticsSchema() -> str:
    return _load_schema("DashboardUsageStatistics")

@functools.lru_cache(maxsize=None)
def getTestInfoSchema() -> str:
    return _load_schema("TestInfo")

@functools.lru_cache(maxsize=None)
def getTestResultsSchema() -> str:
    return _load_schema("TestResults")

@functools.lru_cache(maxsize=None)
def getMLHyperParamSchema() -> str:
    return _load_schema("MLHyperParam")

@functools.lru_cache(maxsize=None)
def getIntendedUseSchema() -> str:
    return _load_schema("IntendedUse")

@functools.lru_cache(maxsize=None)
def getMLModelDeploymentPropertiesSchema() -> str:
    return _load_schema("MLModelDeploymentProperties")

@functools.lru_cache(maxsize=None)
def getMLModelGroupPropertiesSchema() -> str:
    return _load_schema("MLModelGroupProperties")

@functools.lru_cache(maxsize=None)
def getMLMetricSchema() -> str:
    return _load_schema("MLMetric")

@functools.lru_cache(maxsize=None)
def getQuantitativeAnalysesSchema() -> str:
    return _load_schema("QuantitativeAnalyses")

@functools.lru_cache(maxsize=None)
def getEvaluationDataSchema() -> str:
    return _load_schema("EvaluationData")

@functools.lru_cache(maxsize=None)
def getEthicalConsiderationsSchema() -> str:
    return _load_schema("EthicalConsiderations")

@functools.lru_cache(maxsize=None)
def getEditableMLFeatureTablePropertiesSchema() -> str:
    return _load_schema("EditableMLFeatureTableProperties")

@functools.lru_cache(maxsize=None)
def getEditableMLModelGroupPropertiesSchema() -> str:
    return _load_schema("EditableMLModelGroupProperties")

@functools.lru_cache(maxsize=None)
def getMLFeatureTablePropertiesSchema() -> str:
    return _load_schema("MLFeatureTableProperties")

@functools.lru_cache(maxsize=None)
def getEditableMLPrimaryKeyPropertiesSchema() -> str:
    return _load_schema("EditableMLPrimaryKeyProperties")

@functools.lru_cache(maxsize=None)
def getEditableMLModelPropertiesSchema() -> str:
    return _load_schema("EditableMLModelProperties")

@functools.lru_cache(maxsize=None)
def getMLModelFactorPromptsSchema() -> str:
    return _load_schema("MLModelFactorPrompts")

@functools.lru_cache(maxsize=None)
def getMetricsSchema() -> str:
    return _load_schema("Metrics")

@functools.lru_cache(maxsize=None)
def getEditableMLFeaturePropertiesSchema() -> str:
    return _load_schema("EditableMLFeatureProperties")

@functools.lru_cache(maxsize=None)
def getMLFeaturePropertiesSchema() -> str:
    return _load_schema("MLFeatureProperties")

@functools.lru_cache(maxsize=None)
def getCaveatsAndRecommendationsSchema() -> str:
    return _load_schema("CaveatsAndRecommendations")

@functools.lru_cache(maxsize=None)
def getTrainingDataSchema() -> str:
    return _load_schema("TrainingData")

@functools.lru_cache(maxsize=None)
def getSourceCodeSchema() -> str:
    return _load_schema("SourceCode")

@functools.lru_cache(maxsize=None)
def getMLPrimaryKeyPropertiesSchema() -> str:
    return _load_schema("MLPrimaryKeyProperties")

@functools.lru_cache(maxsize=None)
def getMLModelPropertiesSchema() -> str:
    return _load_schema("MLModelProperties")

@functools.lru_cache(maxsize=None)
def getGlossaryTermsSchema() -> str:
    return _load_schema("GlossaryTerms")

@functools.lru_cache(maxsize=None)
def getStatusSchema() -> str:
    return _load_schema("Status")

@functools.lru_cache(maxsize=None)
def getInputFieldsSchema() -> str:
    return _load_schema("InputFields")

@functools.lru_cache(maxsize=None)
def getCostSchema() -> str:
    return _load_schema("Cost")

@functools.lru_cache(maxsize=None)
def getDataPlatformInstanceSchema() -> str:
    return _load_schema("DataPlatformInstance")

@functools.lru_cache(maxsize=None)
def getBrowsePathsSchema() -> str:
    return _load_schema("BrowsePaths")

@functools.lru_cache(maxsize=None)
def getOriginSchema() -> str:
    return _load_schema("Origin")

@functools.lru_cache(maxsize=None)
def getOperationSchema() -> str:
    return _load_schema("Operation")

@functools.lru_cache(maxsize=None)
def getSubTypesSchema() -> str:
    return _load_schema("SubTypes")

@functools.lru_cache(maxsize=None)
def getOwnershipSchema() -> str:
    return _load_schema("Ownership")

@functools.lru_cache(maxsize=None)
def getEmbedSchema() -> str:
    return _load_schema("Embed")

@functools.lru_cache(maxsize=None)
def getSiblingsSchema() -> str:
    return _load_schema("Siblings")

@functools.lru_cache(maxsize=None)
def getInstitutionalMemorySchema() -> str:
    return _load_schema("InstitutionalMemory")

@functools.lru_cache(maxsize=None)
def getDeprecationSchema() -> str:
    return _load_schema("Deprecation")

@functools.lru_cache(maxsize=None)
def getGlobalTagsSchema() -> str:
    return _load_schema("GlobalTags")

@functools.lru_cache(maxsize=None)
def getDataPlatformInfoSchema() -> str:
    return _load_schema("DataPlatformInfo")

@functools.lru_cache(maxsize=None)
def getEntityChangeEventSchema() -> str:
    return _load_schema("EntityChangeEvent")

@functools.lru_cache(maxsize=None)
def getEditableContainerPropertiesSchema() -> str:
    return _load_schema("EditableContainerProperties")

@functools.lru_cache(maxsize=None)
def getContainerSchema() -> str:
    return _load_schema("Container")

@functools.lru_cache(maxsize=None)
def getContainerPropertiesSchema() -> str:
    return _load_schema("ContainerProperties")

@functools.lru_cache(maxsize=None)
def getGlobalSettingsInfoSchema() -> str:
    return _load_schema("GlobalSettingsInfo")

@functools.lru_cache(maxsize=None)
def getDataHubRetentionConfigSchema() -> str:
    return _load_schema("DataHubRetentionConfig")

@functools.lru_cache(maxsize=None)
def getDomainsSchema() -> str:
    return _load_schema("Domains")

@functools.lru_cache(maxsize=None)
def getDomainPropertiesSchema() -> str:
    return _load_schema("DomainProperties")

@functools.lru_cache(maxsize=None)
def getDataHubIngestionSourceInfoSchema() -> str:
    return _load_schema("DataHubIngestionSourceInfo")

@functools.lru_cache(maxsize=None)
def getEditableSchemaMetadataSchema() -> str:
    return _load_schema("EditableSchemaMetadata")

@functools.lru_cache(maxsize=None)
def getSchemaMetadataSchema() -> str:
    return _load_schema("SchemaMetadata")

@functools.lru_cache(maxsize=None)
def getFilterSchema() -> str:
    return _load_schema("Filter")

@functools.lru_cache(maxsize=None)
def getGlobalSettingsKeySchema() -> str:
    return _load_schema("GlobalSettingsKey")

@functools.lru_cache(maxsize=None)
def getPostKeySchema() -> str:
    return _load_schema("PostKey")

@functools.lru_cache(maxsize=None)
def getTagKeySchema() -> str:
    return _load_schema("TagKey")

@functools.lru_cache(maxsize=None)
def getDataPlatformInstanceKeySchema() -> str:
    return _load_schema("DataPlatformInstanceKey")

@functools.lru_cache(maxsize=None)
def getDataHubSecretKeySchema() -> str:
    return _load_schema("DataHubSecretKey")

@functools.lru_cache(maxsize=None)
def getMLFeatureKeySchema() -> str:
    return _load_schema("MLFeatureKey")

@functools.lru_cache(maxsize=None)
def getDataJobKeySchema() -> str:
    return _load_schema("DataJobKey")

@functools.lru_cache(maxsize=None)
def getGlossaryNodeKeySchema() -> str:
    return _load_schema("GlossaryNodeKey")

@functools.lru_cache(maxsize=None)
def getDataProcessKeySchema() -> str:
    return _load_schema("DataProcessKey")

@functools.lru_cache(maxsize=None)
def getCorpUserKeySchema() -> str:
    return _load_schema("CorpUserKey")

@functools.lru_cache(maxsize=None)
def getDataHubViewKeySchema() -> str:
    return _load_schema("DataHubViewKey")

@functools.lru_cache(maxsize=None)
def getDataHubRoleKeySchema() -> str:
    return _load_schema("DataHubRoleKey")

@functools.lru_cache(maxsize=None)
def getDataHubPolicyKeySchema() -> str:
    return _load_schema("DataHubPolicyKey")

@functools.lru_cache(maxsize=None)
def getMLModelKeySchema() -> str:
    return _load_schema("MLModelKey")

@functools.lru_cache(maxsize=None)
def getMLFeatureTableKeySchema() -> str:
    return _load_schema("MLFeatureTableKey")

@functools.lru_cache(maxsize=None)
def getTelemetryKeySchema() -> str:
    return _load_schema("TelemetryKey")

@functools.lru_cache(maxsize=None)
def getChartKeySchema() -> str:
    return _load_schema("ChartKey")

@functools.lru_cache(maxsize=None)
def getQueryKeySchema() -> str:
    return _load_schema("QueryKey")

@functools.lru_cache(maxsize=None)
def getDataHubUpgradeKeySchema() -> str:
    return _load_schema("DataHubUpgradeKey")

@functools.lru_cache(maxsize=None)
def getDataHubRetentionKeySchema() -> str:
    return _load_schema("DataHubRetentionKey")

@functools.lru_cache(maxsize=None)
def getDatasetKeySchema() -> str:
    return _load_schema("DatasetKey")

@functools.lru_cache(maxsize=None)
def getTestKeySchema() -> str:
    return _load_schema("TestKey")

@functools.lru_cache(maxsize=None)
def getMLModelGroupKeySchema() -> str:
    return _load_schema("MLModelGroupKey")

@functools.lru_cache(maxsize=None)
def getDataHubIngestionSourceKeySchema() -> str:
    return _load_schema("DataHubIngestionSourceKey")

@functools.lru_cache(maxsize=None)
def getMLModelDeploymentKeySchema() -> str:
    return _load_schema("MLModelDeploymentKey")

@functools.lru_cache(maxsize=None)
def getInviteTokenKeySchema() -> str:
    return _load_schema("InviteTokenKey")

@functools.lru_cache(maxsize=None)
def getDomainKeySchema() -> str:
    return _load_schema("DomainKey")

@functools.lru_cache(maxsize=None)
def getDataPlatformKeySchema() -> str:
    return _load_schema("DataPlatformKey")

@functools.lru_cache(maxsize=None)
def getContainerKeySchema() -> str:
    return _load_schema("ContainerKey")

@functools.lru_cache(maxsize=None)
def getDataFlowKeySchema() -> str:
    return _load_schema("DataFlowKey")

@functools.lru_cache(maxsize=None)
def getAssertionKeySchema() -> str:
    return _load_schema("AssertionKey")

@functools.lru_cache(maxsize=None)
def getGlossaryTermKeySchema() -> str:
    return _load_schema("GlossaryTermKey")

@functools.lru_cache(maxsize=None)
def getExecutionRequestKeySchema() -> str:
    return _load_schema("ExecutionRequestKey")

@functools.lru_cache(maxsize=None)
def getSchemaFieldKeySchema() -> str:
    return _load_schema("SchemaFieldKey")

@functools.lru_cache(maxsize=None)
def getDataHubStepStateKeySchema() -> str:
    return _load_schema("DataHubStepStateKey")

@functools.lru_cache(maxsize=None)
def getMLPrimaryKeyKeySchema() -> str:
    return _load_schema("MLPrimaryKeyKey")

@functools.lru_cache(maxsize=None)
def getCorpGroupKeySchema() -> str:
    return _load_schema("CorpGroupKey")

@functools.lru_cache(maxsize=None)
def getNotebookKeySchema() -> str:
    return _load_schema("NotebookKey")

@functools.lru_cache(maxsize=None)
def getDashboardKeySchema() -> str:
    return _load_schema("DashboardKey")

@functools.lru_cache(maxsize=None)
def getDataHubAccessTokenKeySchema() -> str:
    return _load_schema("DataHubAccessTokenKey")

@functools.lru_cache(maxsize=None)
def getDataProcessInstanceKeySchema() -> str:
    return _load_schema("DataProcessInstanceKey")

@functools.lru_cache(maxsize=None)
def getUsageAggregationSchema() -> str:
    return _load_schema("UsageAggregation")

@functools.lru_cache(maxsize=None)
def getDataHubStepStatePropertiesSchema() -> str:
    return _load_schema("DataHubStepStateProperties")

@functools.lru_cache(maxsize=None)
def getTagPropertiesSchema() -> str:
    return _load_schema("TagProperties")

@functools.lru_cache(maxsize=None)
def getDataHubAccessTokenInfoSchema() -> str:
    return _load_schema("DataHubAccessTokenInfo")

@functools.lru_cache(maxsize=None)
def getEditableDataFlowPropertiesSchema() -> str:
    return _load_schema("EditableDataFlowProperties")

@functools.lru_cache(maxsize=None)
def getEditableDataJobPropertiesSchema() -> str:
    return _load_schema("EditableDataJobProperties")

@functools.lru_cache(maxsize=None)
def getVersionInfoSchema() -> str:
    return _load_schema("VersionInfo")

@functools.lru_cache(maxsize=None)
def getDataJobInputOutputSchema() -> str:
    return _load_schema("DataJobInputOutput")

@functools.lru_cache(maxsize=None)
def getDataJobInfoSchema() -> str:
    return _load_schema("DataJobInfo")

@functools.lru_cache(maxsize=None)
def getDataFlowInfoSchema() -> str:
    return _load_schema("DataFlowInfo")

@functools.lru_cache(maxsize=None)
def getDatahubIngestionRunSummarySchema() -> str:
    return _load_schema("DatahubIngestionRunSummary")

@functools.lru_cache(maxsize=None)
def getDatahubIngestionCheckpointSchema() -> str:
    return _load_schema("DatahubIngestionCheckpoint")

@functools.lru_cache(maxsize=None)
def getDataHubSecretValueSchema() -> str:
    return _load_schema("DataHubSecretValue")

@functools.lru_cache(maxsize=None)
def getCorpUserSettingsSchema() -> str:
    return _load_schema("CorpUserSettings")

@functools.lru_cache(maxsize=None)
def getCorpUserStatusSchema() -> str:
    return _load_schema("CorpUserStatus")

@functools.lru_cache(maxsize=None)
def getRoleMembershipSchema() -> str:
    return _load_schema("RoleMembership")

@functools.lru_cache(maxsize=None)
def getCorpUserCredentialsSchema() -> str:
    return _load_schema("CorpUserCredentials")

@functools.lru_cache(maxsize=None)
def getCorpGroupInfoSchema() -> str:
    return _load_schema("CorpGroupInfo")

@functools.lru_cache(maxsize=None)
def getInviteTokenSchema() -> str:
    return _load_schema("InviteToken")

@functools.lru_cache(maxsize=None)
def getGroupMembershipSchema() -> str:
    return _load_schema("GroupMembership")

@functools.lru_cache(maxsize=None)
def getCorpGroupEditableInfoSchema() -> str:
    return _load_schema("CorpGroupEditableInfo")

@functools.lru_cache(maxsize=None)
def getCorpUserEditableInfoSchema() -> str:
    return _load_schema("CorpUserEditableInfo")

@functools.lru_cache(maxsize=None)
def getNativeGroupMembershipSchema() -> str:
    return _load_schema("NativeGroupMembership")

@functools.lru_cache(maxsize=None)
def getCorpUserInfoSchema() -> str:
    return _load_schema("CorpUserInfo")

@functools.lru_cache(maxsize=None)
def getDataHubViewInfoSchema() -> str:
    return _load_schema("DataHubViewInfo")

@functools.lru_cache(maxsize=None)
def getChartInfoSchema() -> str:
    return _load_schema("ChartInfo")

@functools.lru_cache(maxsize=None)
def getChartQuerySchema() -> str:
    return _load_schema("ChartQuery")

@functools.lru_cache(maxsize=None)
def getEditableChartPropertiesSchema() -> str:
    return _load_schema("EditableChartProperties")

@functools.lru_cache(maxsize=None)
def getChartUsageStatisticsSchema() -> str:
    return _load_schema("ChartUsageStatistics")

@functools.lru_cache(maxsize=None)
def getQueryPropertiesSchema() -> str:
    return _load_schema("QueryProperties")

@functools.lru_cache(maxsize=None)
def getQuerySubjectsSchema() -> str:
    return _load_schema("QuerySubjects")

@functools.lru_cache(maxsize=None)
def getDataHubRoleInfoSchema() -> str:
    return _load_schema("DataHubRoleInfo")

@functools.lru_cache(maxsize=None)
def getDataHubPolicyInfoSchema() -> str:
    return _load_schema("DataHubPolicyInfo")

@functools.lru_cache(maxsize=None)
def getEditableNotebookPropertiesSchema() -> str:
    return _load_schema("EditableNotebookProperties")

@functools.lru_cache(maxsize=None)
def getNotebookInfoSchema() -> str:
    return _load_schema("NotebookInfo")

@functools.lru_cache(maxsize=None)
def getNotebookContentSchema() -> str:
    return _load_schema("NotebookContent")

@functools.lru_cache(maxsize=None)
def getDataPlatformInstancePropertiesSchema() -> str:
    return _load_schema("DataPlatformInstanceProperties")

@functools.lru_cache(maxsize=None)
def getPostInfoSchema() -> str:
    return _load_schema("PostInfo")

@functools.lru_cache(maxsize=None)
def getDataHubUpgradeResultSchema() -> str:
    return _load_schema("DataHubUpgradeResult")

@functools.lru_cache(maxsize=None)
def getDataHubUpgradeRequestSchema() -> str:
    return _load_schema("DataHubUpgradeRequest")

# fmt: on
